cdr6_ancestry_pred_path = "gs://fc-aou-datasets-controlled/v6/wgs/vcf/aux/ancestry/ancestry_preds.tsv"
cdr7_ancestry_pred_path = "gs://fc-aou-datasets-controlled/v7/wgs/short_read/snpindel/aux/ancestry/ancestry_preds.tsv"

cdr6_mt_path = "gs://fc-aou-datasets-controlled/v6/wgs/hail.mt"
cdr7_mt_path = "gs://fc-aou-datasets-controlled/v7/wgs/short_read/snpindel/acaf_threshold_v7.1/multiMT/hail.mt"

# obsolete paths
# cdr7_mt_path = "gs://fc-aou-datasets-controlled/v7/wgs/short_read/snpindel/acaf_threshold/multiMT/hail.mt"
